type exp = Int of int | Var of string 
       | Plus of exp * exp | Times of exp * exp
type stmt = Skip | Assign of string * exp
         | Seq of stmt * stmt 
         | If of exp * stmt * stmt
         | While of exp * stmt
type heap = (string * int) list

let rec lookup h str =
  match h with
    [] -> 0 (* kind of a cheat *)
  |(s,i)::tl -> if s=str then i else lookup tl str

let update h str i = (str,i)::h

exception AlreadyValue
let rec interp_e (h:heap) (e:exp) =
 match e with
  Int i   -> raise AlreadyValue
 |Var str -> Int (lookup h str)
 |Plus(Int i1,Int i2) -> Int (i1+i2)
 |Plus(Int i1, e2) -> Plus(Int i1,interp_e h e2)
 |Plus(e1, e2) -> Plus(interp_e h e1,e2)
 |Times(Int i1,Int i2) -> Int (i1*i2)
 |Times(Int i1, e2)-> Times(Int i1,interp_e h e2)
 |Times(e1, e2) -> Times(interp_e h e1,e2)

let rec interp_s (h:heap) (s:stmt) =
 match s with
  Skip             -> raise AlreadyValue
 |Assign(str,Int i)-> ((update h str i),Skip)
 |Assign(str,e)    -> (h,Assign(str,interp_e h e))
 |Seq(Skip,s2)     -> (h,s2)
 |Seq(s1,s2)       -> let (h2,s3) = interp_s h s1
                      in (h2,Seq(s3,s2))
 |If(Int i,s1,s2)  -> (h, if i <> 0 
                          then s1
                          else s2)
 |If(e,s1,s2) -> (h, If(interp_e h e, s1, s2))
 |While(e,s1) -> (h, If(e,Seq(s1,s),Skip))

let mt_heap = [] (* common PL pun *)
let interp_prog s = 
  let rec loop (h,s) =
     match s with
       Skip -> lookup h "ans"
     | _    -> loop (interp_s h s)
  in loop (mt_heap,s)
